package gssim.db.analyzer;

import gssim.db.config.Config;
import gssim.db.config.MySQLConfig;
import gssim.db.dao.data.Allocation;
import gssim.db.dao.data.BlobProperty;
import gssim.db.dao.data.Configuration;
import gssim.db.dao.data.Experiment;
import gssim.db.dao.data.ExperimentStats;
import gssim.db.dao.data.Queues;
import gssim.db.dao.data.Reservation;
import gssim.db.dao.data.ReservationDetails;
import gssim.db.dao.data.Resource;
import gssim.db.dao.data.ResourceAllTimeline;
import gssim.db.dao.data.ResourceEnTimeline;
import gssim.db.dao.data.Task;
import gssim.db.dao.data.TextProperty;
import gssim.db.dao.data.Unit;
import gssim.db.dao.data.UnitDetails;
import gssim.db.dao.data.User;
import gssim.db.mysql.AllocationDAO;
import gssim.db.mysql.BlobPropertyDAO;
import gssim.db.mysql.ConfigurationDAO;
import gssim.db.mysql.ExperimentDAO;
import gssim.db.mysql.ExperimentStatsDAO;
import gssim.db.mysql.MySQLDAOFactory;
import gssim.db.mysql.QueuesDAO;
import gssim.db.mysql.ReservationDAO;
import gssim.db.mysql.ReservationDetailsDAO;
import gssim.db.mysql.ResourceAllTimelineDAO;
import gssim.db.mysql.ResourceDAO;
import gssim.db.mysql.ResourceEnTimelineDAO;
import gssim.db.mysql.TaskDAO;
import gssim.db.mysql.TextPropertyDAO;
import gssim.db.mysql.UnitDAO;
import gssim.db.mysql.UnitDetailsDAO;
import gssim.db.mysql.UserDAO;
import gssim.db.reader.AccResourcesReader;
import gssim.db.reader.AccStatsReader;
import gssim.db.reader.JobsReader;
import gssim.db.reader.NetworkReader;
import gssim.db.reader.RawResourcesReader;
import gssim.db.reader.StatsParser;
import gssim.db.reader.TaskReader;
import gssim.db.reader.UsageReader;
import gssim.db.reader.RawResourcesReader.Load;
import gssim.db.utils.IOUtils;
import gssim.db.utils.Pair;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Map.Entry;

import org.apache.commons.compress.archivers.tar.TarArchiveEntry;
import org.apache.commons.compress.archivers.tar.TarArchiveInputStream;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * The Class Analyzer. It analyzes GSSIM's reuslts text files ant put data from
 * them to the database
 * 
 * @author Jarek Szymczak
 */

public class Analyzer {

	protected Log log = LogFactory.getLog(Analyzer.class);

	/** The job id map. It maps job id from text file to database primary key */
	private Map<Integer, Integer> jobID;

	/**
	 * The task id map. It maps job id and task id from text file to database
	 * primary key.
	 */
	private Map<Pair<Integer, Integer>, Integer> taskID;

	/**
	 * The unit id map. It maps unit name from text file to database primary
	 * key.
	 */
	private Map<String, Integer> unitID;

	/**
	 * The network task id. It maps network reservation id with task database
	 * primary key
	 */
	private Map<Integer, Integer> networkTaskID;

	/**
	 * The resource id map. It maps resource name from text file to database
	 * primary key.
	 */
	private Map<String, Integer> resourceID;

	/** The jobs' start dates. */
	private Map<Integer, Timestamp> jobStartDate;

	/** The jobs' end dates. */
	private Map<Integer, Timestamp> jobEndDate;

	/** The jobs' finish dates. */
	private Map<Integer, Timestamp> jobFinishDate;

	/**
	 * The input streams map. It's used to be source (directory or archive file)
	 * independent
	 */
	private Map<String, InputStream> inputStreams;

	private List<String> files;
	private MySQLDAOFactory factory;

	private User user;
	private Configuration configuration;
	private Experiment experiment;
	private String userDN;
	private Integer totalSimulationRuns;
	private Integer currentSimulationRun;

	public Analyzer() throws InstantiationException,
			IllegalAccessException, ClassNotFoundException {

		jobID = new HashMap<Integer, Integer>();
		taskID = new HashMap<Pair<Integer, Integer>, Integer>();
		unitID = new HashMap<String, Integer>();
		networkTaskID = new HashMap<Integer, Integer>();
		resourceID = new HashMap<String, Integer>();
		jobStartDate = new HashMap<Integer, Timestamp>();
		jobEndDate = new HashMap<Integer, Timestamp>();
		jobFinishDate = new HashMap<Integer, Timestamp>();
		inputStreams = new HashMap<String, InputStream>();
		files = new ArrayList<String>();
		user = null;
		configuration = null;
		experiment = null;
		userDN = null;
		factory = new MySQLDAOFactory();
		totalSimulationRuns = 1;
	}

	/**
	 * The main method.
	 * 
	 * @param args
	 *            the arguments
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 * @throws ClassNotFoundException
	 * @throws IllegalAccessException
	 * @throws InstantiationException
	 */
	public static void main(String[] args) throws IOException,
			InstantiationException, IllegalAccessException,
			ClassNotFoundException {
		Analyzer analyzer = new Analyzer();
		analyzer.run(args);
	}

	/**
	 * Non-static main method.
	 * 
	 * @param args
	 *            arguments passed from main
	 */
	public void run(String[] args) {

		Config config = null;

		if (args.length == 1) {
			config = new MySQLConfig();
		} else if (args.length == 2) {
			try {
				config = new MySQLConfig(args[1]);
			} catch (FileNotFoundException e) {
				log.debug("Properties file wasn't found.", e);
			} catch (IOException e) {
				log.debug("IOException occured", e);
			}
		} else if (args.length == 3) {
			try {
				config = new MySQLConfig(args[1], args[2]);
			} catch (FileNotFoundException e) {
				log.debug("Properties or key file is missing.", e);
			} catch (IOException e) {
				log.debug("IOException occured", e);
			}
		}

		if (args.length < 1 || args.length > 3)
			throw new RuntimeException(
					"Program should be run with 1-3 arguments: "
							+ "*.tar file or directory with experiment results, "
							+ "encrypted or unencrypted properties file, "
							+ "key file for encrypted properties");

		File gssimResults = new File(args[0]);
		TarArchiveInputStream tarInputStream = null;
		Connection connection = null;

		try {

			connection = factory.createDBConnection(config.getDb(), config
					.getUser(), config.getPassword());

			connection.setAutoCommit(false);

			currentSimulationRun = 0;
			
			while (currentSimulationRun != totalSimulationRuns)
			{
				currentSimulationRun++;
				config.initFileNames(currentSimulationRun);
				if (gssimResults.getName().endsWith(".tar")) {
	
					FileInputStream fileStream = new FileInputStream(gssimResults);
					tarInputStream = new TarArchiveInputStream(fileStream);
					prepareStreams(tarInputStream, config);
	
				} else if (gssimResults.isDirectory()) {
	
					PrepareStreams(gssimResults, config);
	
				} else
					throw new RuntimeException(
							"This program should be run with one argument: "
									+ "*.tar file or directory with experiment results");
	
				checkFilesPresence(config);
				
				setup(inputStreams.get("propertiesInputStream"), config);
				process();
			}
			connection.commit();
			log.info("Database operations are commited");

		} catch (IOException e) {
			log.error("IO exception occured", e);

		} catch (SQLException e) {
			try {
				log.error("SQL exception occured, operations "
						+ "will be withdrawn (rollback)", e);
				connection.rollback();
			} catch (SQLException e1) {
				log.error("Rollback failed: " + e.toString());
			}
		} catch (Exception e) {
			try {
				log.error("Exception occured, operations "
						+ "will be withdrawn (rollback)", e);
				connection.rollback();
			} catch (SQLException e1) {
				log.error("Rollback failed: " + e.toString());
			}
		} finally {
			if (connection != null) {
				try {
					connection.close();
				} catch (SQLException e) {
					log.error("Error occured while closing the connection: "
							+ e.getMessage());
				}
			}

			if (tarInputStream != null)
				try {
					tarInputStream.close();
				} catch (IOException e) {
					log.error("Exception occured while closing the stream", e);
				}

			for (Entry<String, InputStream> inputStreamEntry : inputStreams
					.entrySet()) {
				if (inputStreamEntry.getValue() != null)
					try {
						inputStreamEntry.getValue().close();
					} catch (IOException e) {
						log.error("Exception occured while closing the "
								+ inputStreamEntry.getKey(), e);
					}
			}
		}

	}

	/**
	 * Prepares streams from *.tar file input stream (through byte arrays in
	 * memory or through files if archive is extracted).
	 * 
	 * @param tarInputStream
	 *            the input stream from *.tar file
	 * @param config
	 *            configuration class object with important informations
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	private void prepareStreams(TarArchiveInputStream tarInputStream,
			Config config) throws IOException {

		String statsFile = config.getStatsFile();
		String jobsFile = config.getJobsFile();
		String tasksFile = config.getTasksFile();
		String resourcesFile = config.getResourcesFile();
		String energyFile = config.getEnergyFile();
		String resourceLoadFile = config.getResourceLoadFile();
		String accumulatedResourcesFile = config.getAccumulatedResourcesFile();
		String networkFile = config.getNetworkFile();
		String propertiesFile = config.getPropertiesFile();

		TarArchiveEntry entry = null;

		String extractDirectoryPath = config.getExtractDirectoryPath();
		if (extractDirectoryPath != null) {
			File extractDirectory = new File(extractDirectoryPath);
			if (extractDirectory.exists()) {
				if (!extractDirectory.isDirectory()) {
					extractDirectory.delete();
					extractDirectory.mkdir();
				}

			} else
				extractDirectory.mkdir();

			while ((entry = tarInputStream.getNextTarEntry()) != null) {
				if (entry.isDirectory())
					continue;
				String name = entry.getName();
				if (name == null)
					continue;
				if (name.endsWith(".sjg") || name.endsWith(".png"))
					continue;
				if (name.contains(File.separator)) {
					int lastSeparatorIndex = name.lastIndexOf(File.separator);
					String dirPath = extractDirectory.getAbsolutePath()
							+ File.separator
							+ name.substring(0, lastSeparatorIndex);
					File dir = new File(dirPath);
					if (dir.exists())
						dir.delete();
					dir.mkdirs();
				}

				OutputStream out = null;
				try {
					out = new FileOutputStream(extractDirectory
							.getAbsolutePath()
							+ File.separator + entry.getName());
					IOUtils.copy(tarInputStream, out);
				} finally {
					if (out != null)
						out.close();
				}
			}

			PrepareStreams(extractDirectory, config);

		} else {

			Map<String, byte[]> inputArrays = new HashMap<String, byte[]>();

			while ((entry = tarInputStream.getNextTarEntry()) != null) {
				if (entry.isDirectory())
					continue;
				String name = entry.getName();
				if (name == null)
					continue;
				if (name.endsWith(".sjg") || name.endsWith(".png"))
					continue;

				files.add(name);
				inputArrays.put(name, IOUtils.toByteArray(tarInputStream));
			}

			for (Entry<String, byte[]> inputArrayEntry : inputArrays.entrySet()) {

				if (statsFile.equals(inputArrayEntry.getKey())) {
					inputStreams
							.put("statsInputStream", new ByteArrayInputStream(
									inputArrayEntry.getValue()));

				} else if (jobsFile.equals(inputArrayEntry.getKey())) {
					inputStreams
							.put("jobsInputStream", new ByteArrayInputStream(
									inputArrayEntry.getValue()));

				} else if (tasksFile.equals(inputArrayEntry.getKey())) {
					inputStreams
							.put("tasksInputStream", new ByteArrayInputStream(
									inputArrayEntry.getValue()));
					inputStreams
							.put("reservationsInputStream",
									new ByteArrayInputStream(inputArrayEntry
											.getValue()));
					inputStreams
							.put("allocationsInputStream",
									new ByteArrayInputStream(inputArrayEntry
											.getValue()));

				} else if (resourcesFile.equals(inputArrayEntry.getKey())) {
					inputStreams
							.put("unitsInputStream", new ByteArrayInputStream(
									inputArrayEntry.getValue()));
					inputStreams
							.put("queuesInputStream", new ByteArrayInputStream(
									inputArrayEntry.getValue()));

				} else if (resourceLoadFile.equals(inputArrayEntry.getKey())) {
					inputStreams
							.put("resourceAllTimelineInputStream",
									new ByteArrayInputStream(inputArrayEntry
											.getValue()));

				} else if (energyFile.equals(inputArrayEntry.getKey())) {
					inputStreams
							.put("resourceEnTimelineInputStream",
									new ByteArrayInputStream(inputArrayEntry
											.getValue()));

				} else if (accumulatedResourcesFile.equals(inputArrayEntry
						.getKey())) {
					inputStreams
							.put("resourcesInputStream",
									new ByteArrayInputStream(inputArrayEntry
											.getValue()));

				} else if (networkFile.equals(inputArrayEntry.getKey())) {
					inputStreams
							.put("networkInputStream",
									new ByteArrayInputStream(inputArrayEntry
											.getValue()));

				} else if (propertiesFile.equals(inputArrayEntry.getKey())) {
					inputStreams
							.put("propertiesInputStream",
									new ByteArrayInputStream(inputArrayEntry
											.getValue()));

				} else
					inputStreams
							.put(inputArrayEntry.getKey(),
									new ByteArrayInputStream(inputArrayEntry
											.getValue()));

			}
		}

	}

	/**
	 * Prepares streams from directory.
	 * 
	 * @param directory
	 *            the directory
	 * @param config
	 *            configuration class object with important informations
	 * @throws FileNotFoundException
	 *             the file not found exception
	 */
	void PrepareStreams(File directory, Config config)
			throws FileNotFoundException {

		if (!directory.isDirectory())
			throw new RuntimeException("Usage of not proper directory file");

		for (String fileName : directory.list()) {
			files.add(fileName);
		}

		inputStreams.put("statsInputStream", new FileInputStream(directory
				+ File.separator + config.getStatsFile()));
		inputStreams.put("jobsInputStream", new FileInputStream(directory
				+ File.separator + config.getJobsFile()));
		inputStreams.put("tasksInputStream", new FileInputStream(directory
				+ File.separator + config.getTasksFile()));
		inputStreams.put("resourcesInputStream", new FileInputStream(directory
				+ File.separator + config.getAccumulatedResourcesFile()));
		inputStreams.put("resourceAllTimelineInputStream", new FileInputStream(
				directory + File.separator + config.getResourceLoadFile()));
		inputStreams.put("resourceEnTimelineInputStream", new FileInputStream(
				directory + File.separator + config.getEnergyFile()));
		inputStreams.put("unitsInputStream", new FileInputStream(directory
				+ File.separator + config.getResourcesFile()));
		inputStreams.put("queuesInputStream", new FileInputStream(directory
				+ File.separator + config.getResourcesFile()));
		inputStreams.put("allocationsInputStream", new FileInputStream(
				directory + File.separator + config.getTasksFile()));
		inputStreams.put("reservationsInputStream", new FileInputStream(
				directory + File.separator + config.getTasksFile()));
		inputStreams.put("propertiesInputStream", new FileInputStream(directory
				+ File.separator + config.getPropertiesFile()));

		File networkFile = null;

		if (config.getNetworkFile() != null)
			networkFile = new File(directory + File.separator
					+ config.getNetworkFile());

		if (networkFile.exists())
			inputStreams.put("networkInputStream", new FileInputStream(
					networkFile));
	}

	/**
	 * Checks whether indispensible files are present.
	 * 
	 * @param config
	 *            configuration class object with important informations
	 */
	private void checkFilesPresence(Config config) {

		if (!files.contains(config.getEnergyFile()))
			throw new RuntimeException("Lack of file (energy): "
					+ config.getEnergyFile());

		if (!files.contains(config.getStatsFile()))
			throw new RuntimeException("Lack of file (stats): "
					+ config.getStatsFile());

		if (!files.contains(config.getJobsFile()))
			throw new RuntimeException("Lack of file (jobs): "
					+ config.getJobsFile());

		if (!files.contains(config.getResourceLoadFile()))
			throw new RuntimeException("Lack of file (resource load): "
					+ config.getResourceLoadFile());

		if (!files.contains(config.getResourcesFile()))
			throw new RuntimeException("Lack of file (resources): "
					+ config.getResourcesFile());

		if (!files.contains(config.getTasksFile()))
			throw new RuntimeException("Lack of file (tasks): "
					+ config.getTasksFile());

		if (!files.contains(config.getPropertiesFile()))
			throw new RuntimeException("Lack of file (tasks): "
					+ config.getPropertiesFile());

		if (!files.contains(config.getPropertiesFile()))
			throw new RuntimeException("Lack of file (tasks): "
					+ config.getPropertiesFile());
	}

	/**
	 * Setup method. It adds configuration and it's properties as well as user
	 * (if new) to the database. It is also responsible for network topology
	 * stream preparation (if it's specified in properties and exists)
	 * 
	 * @param propertiesFileInputStream
	 *            input stream with properties
	 * @param config
	 *            configuration class object with important informations
	 * @throws SQLException
	 *             the SQL exception
	 */
	private void setup(InputStream propertiesFileInputStream, Config config)
			throws SQLException {

		log.debug("Setup begin");

		UserDAO userDAO = factory.createUser();
		ConfigurationDAO configurationDAO = factory.createConfiguration();
		ExperimentDAO experimentDAO = factory.createExperiment();
		TextPropertyDAO textPropertyDAO = factory.createTextProperty();

		Properties properties = new Properties();
		try {
			properties.load(propertiesFileInputStream);
		} catch (IOException e) {
			log.error("Error while loading properties file", e);
		}

		if (properties.containsKey("numberofsimulations"))
		{
			totalSimulationRuns = Integer.parseInt((String) properties.get("numberofsimulations"));
		}
		
		if (properties.containsKey("experiment.id") && currentSimulationRun == 1) {
			experiment = new Experiment();
			experiment.setPK(Integer.parseInt((String) properties.get("experiment.id")));
			experiment = experimentDAO.select(experiment);
			experiment.setStatus("FINISHED");
			experimentDAO.update(experiment, false);
			
			configuration = new Configuration();
			configuration.setId(experiment.getConfigId());
			configuration = configurationDAO.select(configuration);

		} else {

			configuration = new Configuration();

			user = new User();
			user.setDN("unknown_user");

			Integer metaExpId = null;
			String name = "experiment";
			
			if (properties.containsKey("experiment.id"))
			{
				experiment = new Experiment();
				experiment.setPK(Integer.parseInt((String) properties.get("experiment.id")));
				experiment = experimentDAO.select(experiment);
				
				metaExpId = experiment.getMetaexpId();
				name = experiment.getName();
				user.setId(experiment.getUserId());
				configuration.setId(experiment.getConfigId());
			}
			else
			{
				userDAO.insert(user, false);
				configurationDAO.insert(configuration, false);
			}
			
			experiment = new Experiment();
			experiment.setConfigId(configuration.getId());
			experiment.setUserId(user.getId());
			experiment.setCommissionDate(new Timestamp((new Date()).getTime()));
			experiment.setMetaexpId(metaExpId);
			experiment.setName(name+ "_" + currentSimulationRun.toString());
			experiment.setStatus("FINISHED");
			experimentDAO.insert(experiment, false);
		}
		
		for (Map.Entry<Object, Object> entry : properties.entrySet()) {
			if (!config.isIgnoredKey((String) entry.getKey())) {

				TextProperty textProperty = new TextProperty();
				textProperty.setConfigId(configuration.getId());
				textProperty.setName((String) entry.getKey());
				textProperty.setValue((String) entry.getValue());

				textPropertyDAO.insert(textProperty, false);
			}
		}

		File extractDirectory = null;
		String extractDirectoryPath = config.getExtractDirectoryPath();
		if (extractDirectoryPath != null)
			extractDirectory = new File(extractDirectoryPath);
		String networkFilePath = properties
				.getProperty("networktopologyfilename");

		if (inputStreams.get(networkFilePath) != null) {
			inputStreams.put("networkTopologyInputStream", inputStreams
					.get(networkFilePath));

		} else {
			if (networkFilePath != null) {
				File networkFile1 = new File(networkFilePath);

				if (networkFilePath.indexOf(File.separator) != 0)
					networkFilePath = File.separator + networkFilePath;

				File networkFile2 = null;

				if ((extractDirectory != null) && extractDirectory.exists()) {
					networkFile2 = new File(extractDirectory.getAbsolutePath()
							+ networkFilePath);
				}

				if (networkFile1.exists()) {
					try {
						inputStreams.put("networkTopologyInputStream",
								new FileInputStream(networkFile1));
					} catch (FileNotFoundException e) {
						// should not occur
					}

				} else if (networkFile2.exists()) {
					try {
						inputStreams.put("networkTopologyInputStream",
								new FileInputStream(networkFile2));
					} catch (FileNotFoundException e) {
						// should not occur
					}

				} else {
					log
							.error("Error while loading network topology file (it was not found),"
									+ " it will not be loaded to database");
				}

			}
		}

		log.debug("Setup finished");
	}

	/**
	 * Processes input streams in proper order.
	 * 
	 * @throws SQLException
	 *             the SQL exception
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	private void process() throws SQLException, IOException {

		// Order of processing should be preserved
		// (beacause of constraints in database).

		processStats(inputStreams.get("statsInputStream"), true);
		List<Task> jobs = preProcessJobs(inputStreams.get("jobsInputStream"),
				true);
		List<Task> tasks = preProcessTasks(
				inputStreams.get("tasksInputStream"), true);
		processJobs(jobs);
		processTasks(tasks);
		processResources(inputStreams.get("resourcesInputStream"), true);
		processResourceAllTimeline(inputStreams
				.get("resourceAllTimelineInputStream"), true);
		processResourceEnTimeline(inputStreams
				.get("resourceEnTimelineInputStream"), true);
		processUnits(inputStreams.get("unitsInputStream"), true);
		processQueues(inputStreams.get("queuesInputStream"), true);
		processAllocation(inputStreams.get("allocationsInputStream"), true);
		processReservation(inputStreams.get("reservationsInputStream"), true);
		if (inputStreams.get("networkInputStream") != null)
			processNetwork(inputStreams.get("networkInputStream"), true);
		if (inputStreams.get("networkTopologyInputStream") != null)
			processNetworkTopology(inputStreams
					.get("networkTopologyInputStream"));
	}

	/**
	 * Gets the stats parser. It's the basic object for parsing text files with
	 * GSSIM's results.
	 * 
	 * @param is
	 *            the input stream
	 * @param skipHeaders
	 *            determines whether first line of input stream should be
	 *            skipped
	 * @return the stats parser
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	private StatsParser getStatsParser(InputStream is, boolean skipHeaders)
			throws IOException {

		InputStreamReader reader = null;
		reader = new InputStreamReader(is);
		StatsParser statsParser;
		statsParser = new StatsParser(reader, skipHeaders);
		return statsParser;
	}

	/**
	 * Process stats. It reads stats data from the stream and adds them to the
	 * database.
	 * 
	 * @param is
	 *            input stream with stats data
	 * @param skipHeaders
	 *            determines whether first line of input stream should be
	 *            skipped
	 * @throws SQLException
	 *             the SQL exception
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	private void processStats(InputStream is, boolean skipHeaders)
			throws SQLException, IOException {

		log.debug("Processing stats");

		ExperimentStatsDAO statsDAO = factory.createExperimentStats();
		ExperimentStats stats = null;
		List<ExperimentStats> lstats = new ArrayList<ExperimentStats>();
		AccStatsReader statsReader = new AccStatsReader(getStatsParser(is,
				skipHeaders));

		while (statsReader.next()) {

			String metricName = statsReader.getMetricName();

			stats = new ExperimentStats();
			stats.setName(metricName + " mean");
			stats.setValue(statsReader.getMean());
			stats.setId(experiment.getPK());
			lstats.add(stats);

			stats = new ExperimentStats();
			stats.setName(metricName + " stdev");
			stats.setValue(statsReader.getStandardDeviation());
			stats.setId(experiment.getPK());
			lstats.add(stats);

			stats = new ExperimentStats();
			stats.setName(metricName + " variance");
			stats.setValue(statsReader.getVariance());
			stats.setId(experiment.getPK());
			lstats.add(stats);

			stats = new ExperimentStats();
			stats.setName(metricName + " min");
			stats.setValue(statsReader.getMin());
			stats.setId(experiment.getPK());
			lstats.add(stats);

			stats = new ExperimentStats();
			stats.setName(metricName + " max");
			stats.setValue(statsReader.getMax());
			stats.setId(experiment.getPK());
			lstats.add(stats);

			stats = new ExperimentStats();
			stats.setName(metricName + " count");
			stats.setValue(statsReader.getCount());
			stats.setId(experiment.getPK());
			lstats.add(stats);

			stats = new ExperimentStats();
			stats.setName(metricName + " sum");
			stats.setValue(statsReader.getSum());
			stats.setId(experiment.getPK());
			lstats.add(stats);

		}

		statsDAO.insertBatch(lstats, false);

		log.debug("Processing stats finished");
	}

	/**
	 * Pre process jobs.
	 * 
	 * @param is
	 *            input stream with jobs data
	 * @param skipHeaders
	 *            determines whether first line of input stream should be
	 *            skipped
	 * @return the list with jobs without start, end and finish dates
	 * @throws SQLException
	 *             the SQL exception
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	private List<Task> preProcessJobs(InputStream is, boolean skipHeaders)
			throws SQLException, IOException {

		log.debug("Preprocessing jobs");

		List<Task> result = new ArrayList<Task>();

		Task job = null;

		JobsReader jobsReader = new JobsReader(getStatsParser(is, skipHeaders));

		while (jobsReader.next()) {

			job = new Task();

			job.setExpId(experiment.getPK());
			job.setJobId(jobsReader.getJobID());
			job.setLateness(jobsReader.getLateness());
			job.setMakespan(jobsReader.getMakespan());
			job.setCompletionTime(jobsReader.getMeanTaskCompletionTime());
			job.setExecutionTime(jobsReader.getMeanTaskExecutionTime());
			job.setFlowTime(jobsReader.getMeanTaskFlowTime());
			job.setGqWaitingTime(jobsReader.getMeanTaskGQ_WaitingTime());
			job.setStartTime(jobsReader.getMeanTaskStartTime());
			job.setWaitingTime(jobsReader.getMeanTaskWaitingTime());
			job.setTardiness(jobsReader.getTardiness());

			result.add(job);
		}

		log.debug("Preprocessing jobs finished");

		return result;
	}

	/**
	 * Pre process tasks.
	 * 
	 * @param is
	 *            input stream with tasks data
	 * @param skipHeaders
	 *            determines whether first line of input stream should be
	 *            skipped
	 * @return the list with tasks without parent foreign key
	 * @throws SQLException
	 *             the SQL exception
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	private List<Task> preProcessTasks(InputStream is, boolean skipHeaders)
			throws SQLException, IOException {
		log.debug("Preprocessing tasks");

		List<Task> result = new ArrayList<Task>();

		Task task = null;

		TaskReader taskReader = new TaskReader(getStatsParser(is, skipHeaders));

		while (taskReader.next()) {

			task = new Task();

			task.setExpId(experiment.getPK());
			task.setJobId(taskReader.getJobID());
			task.setTaskId(taskReader.getTaskID());
			task.setExecStartDate(taskReader.getExecStartDate());
			task.setExecFinishDate(taskReader.getExecFinishDate());
			task.setExecEndDate(taskReader.getExecEndDate());
			task.setGbSubmitDate(taskReader.getGB_SubDate());
			task.setLbSubmitDate(taskReader.getLB_SubDate());
			task.setCompletionTime(taskReader.getCompletionTime());
			task.setExecStartTime(taskReader.getExecStartTime());
			task.setExecutionTime(taskReader.getExecutionTime());
			task.setStartTime(taskReader.getStartTime());
			task.setReadyTime(taskReader.getReadyTime());
			task.setFlowTime(taskReader.getFlowTime());

			task.setWaitingTime(taskReader.getWaitingTime());
			task.setGqWaitingTime(taskReader.getGQ_WaitTime());
			task.setLateness(taskReader.getLateness());
			task.setTardiness(taskReader.getTardiness());

			manageDates(task);
			result.add(task);
		}

		log.debug("Preprocessing tasks finished");

		return result;
	}

	/**
	 * Manage dates. It fills task with dates and put minimum / maximum of dates
	 * to hash maps. They will be used afterwards to fill jobs dates.
	 * 
	 * @param task
	 *            the task
	 */
	private void manageDates(Task task) {
		Integer jobID = task.getJobId();
		Timestamp startDate = task.getExecStartDate();
		Timestamp endDate = task.getExecEndDate();
		Timestamp finishDate = task.getExecFinishDate();

		if (startDate == null)
			throw new NullPointerException("Lack of start date in task: "
					+ task);
		if (endDate == null)
			throw new NullPointerException("Lack of end date in task: " + task);
		if (finishDate == null)
			throw new NullPointerException("Lack of finish date in task: "
					+ task);

		if (!jobStartDate.containsKey(jobID)
				|| (jobStartDate.get(jobID).compareTo(startDate) > 0))
			jobStartDate.put(jobID, startDate);

		if (!jobEndDate.containsKey(jobID)
				|| (jobEndDate.get(jobID).compareTo(endDate) < 0))
			jobEndDate.put(jobID, endDate);

		if (!jobFinishDate.containsKey(jobID)
				|| (jobFinishDate.get(jobID).compareTo(finishDate) < 0))
			jobFinishDate.put(jobID, finishDate);
	}

	/**
	 * Process jobs. It adds start, finish and end date and put jobs in the
	 * database.
	 * 
	 * @param jobs
	 *            the jobs
	 * @throws SQLException
	 *             the SQL exception
	 */
	private void processJobs(List<Task> jobs) throws SQLException {

		log.debug("Processing jobs");

		TaskDAO taskDAO = factory.createTask();

		for (Task job : jobs) {
			job.setExecStartDate(jobStartDate.get(job.getJobId()));
			job.setExecEndDate(jobEndDate.get(job.getJobId()));
			job.setExecFinishDate(jobFinishDate.get(job.getJobId()));
		}

		taskDAO.insertBatch(jobs, false);

		for (Task job : jobs) {
			jobID.put(job.getJobId(), job.getPK());
		}

		log.debug("Processing jobs finished");
	}

	/**
	 * Process tasks. It adds tasks parent id and puts them into the database.
	 * 
	 * @param tasks
	 *            the tasks
	 * @throws SQLException
	 *             the SQL exception
	 */
	private void processTasks(List<Task> tasks) throws SQLException {

		log.debug("Processing tasks");

		TaskDAO taskDAO = factory.createTask();

		for (Task task : tasks) {
			task.setParentId(jobID.get(task.getJobId()));
		}

		taskDAO.insertBatch(tasks, false);

		for (Task task : tasks) {
			taskID.put(new Pair<Integer, Integer>(task.getJobId(), task
					.getTaskId()), task.getPK());
		}

		log.debug("Processing tasks finished");
	}

	/**
	 * Process resources. It reads resources data from the stream and adds them
	 * to the database.
	 * 
	 * @param is
	 *            input stream with resources data
	 * @param skipHeaders
	 *            determines whether first line of input stream should be
	 *            skipped
	 * @throws SQLException
	 *             the SQL exception
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	private void processResources(InputStream is, boolean skipHeaders)
			throws SQLException, IOException {

		log.debug("Processing resources");

		ResourceDAO resourceDAO = factory.createResource();

		Resource resource = null;

		List<Resource> resources = new ArrayList<Resource>();

		AccResourcesReader resourceReader = new AccResourcesReader(
				getStatsParser(is, skipHeaders));

		for (boolean firstRow = true; resourceReader.next(); firstRow = !firstRow) {

			if (firstRow) {
				resource = new Resource();
				resource.setExpId(experiment.getPK());
				resource.setName(resourceReader.getResourceName());
				resource.setType("type");
				resource.setAllocationLoad(resourceReader.getMean());
			} else {
				resource.setReservationLoad(resourceReader.getMean());
				resources.add(resource);
			}
		}

		resourceDAO.insertBatch(resources, false);

		for (Resource res : resources)
			resourceID.put(res.getName(), res.getPK());

		log.debug("Processing resources finished");
	}

	/**
	 * Process units. It reads units data from the stream and adds them to the
	 * database.
	 * 
	 * @param is
	 *            input stream with units data
	 * @param skipHeaders
	 *            determines whether first line of input stream should be
	 *            skipped
	 * @throws SQLException
	 *             the SQL exception
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	private void processUnits(InputStream is, boolean skipHeaders)
			throws SQLException, IOException {

		log.debug("Processing units");

		UnitDAO unitDAO = factory.createUnit();

		RawResourcesReader unitReader = new RawResourcesReader(getStatsParser(
				is, skipHeaders));

		Map<String, Unit> units = new HashMap<String, Unit>();

		while (unitReader.next()) {

			Load[] cpuLoad = unitReader.getCpuLoad();
			Load[] reservationLoad = unitReader.getReservationLoad();

			for (Load load : cpuLoad) {
				Unit unit = new Unit();
				unit.setAllocationLoad(load.value);
				unit.setAmount(1);
				unit
						.setResourceId(resourceID.get(unitReader
								.getResourceName()));
				unit.setType("type");
				unit.setName(load.cpuName);
				units.put(unit.getName(), unit);
			}

			for (Load load : reservationLoad) {
				Unit unit = units.get(load.cpuName);
				unit.setReservationLoad(load.value);
			}
		}

		unitDAO.insertBatch(new ArrayList<Unit>(units.values()), false);

		for (Unit unit : units.values()) {
			unitID.put(unit.getName(), unit.getPK());
		}

		log.debug("Processing units finished");
	}

	/**
	 * Process allocation. It reads allocation data from the stream and adds
	 * them to the database.
	 * 
	 * @param is
	 *            input stream with allocation data
	 * @param skipHeaders
	 *            determines whether first line of input stream should be
	 *            skipped
	 * @throws SQLException
	 *             the SQL exception
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	private void processAllocation(InputStream is, boolean skipHeaders)
			throws SQLException, IOException {

		log.debug("Processing allocations");

		AllocationDAO allocationDAO = factory.createAllocation();
		Allocation allocation = null;
		List<Allocation> allocations = new ArrayList<Allocation>();
		TaskReader allocationReader = new TaskReader(getStatsParser(is,
				skipHeaders));

		while (allocationReader.next()) {

			String unitNamesString = allocationReader.getProcessorName();
			int endIndex = Math.max(1, unitNamesString.length() - 1);
			unitNamesString = unitNamesString.substring(1, endIndex);
			String[] unitNames = unitNamesString.split(", ");

			for (String unitName : unitNames) {

				allocation = new Allocation();

				allocation.setTaskId(taskID.get(new Pair<Integer, Integer>(
						allocationReader.getJobID(), allocationReader
								.getTaskID())));

				allocation.setUnitId(unitID.get(unitName));
				allocation.setBegin(allocationReader.getExecStartDate());
				allocation.setEnd(allocationReader.getExecFinishDate());

				allocations.add(allocation);
			}
		}

		allocationDAO.insertBatch(allocations, false);

		log.debug("Processing allocaltions finished");
	}

	/**
	 * Process reservation. It reads reservation data from the stream and adds
	 * them to the database.
	 * 
	 * @param is
	 *            input stream with reservation data
	 * @param skipHeaders
	 *            determines whether first line of input stream should be
	 *            skipped
	 * @throws SQLException
	 *             the SQL exception
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	private void processReservation(InputStream is, boolean skipHeaders)
			throws SQLException, IOException {

		log.debug("Processing reservations");

		ReservationDAO reservationDAO = factory.createReservation();
		Reservation reservation = null;
		List<Reservation> reservations = new ArrayList<Reservation>();
		TaskReader reservationReader = new TaskReader(getStatsParser(is,
				skipHeaders));

		while (reservationReader.next()) {

			String unitNamesString = reservationReader.getProcessorName();
			int endIndex = Math.max(1, unitNamesString.length() - 1);
			unitNamesString = unitNamesString.substring(1, endIndex);
			String[] unitNames = unitNamesString.split(", ");

			for (String unitName : unitNames) {

				reservation = new Reservation();

				reservation.setTaskId(taskID.get(new Pair<Integer, Integer>(
						reservationReader.getJobID(), reservationReader
								.getTaskID())));

				reservation.setUnitId(unitID.get(unitName));
				reservation.setBegin(reservationReader.getReservStartDate());
				reservation.setEnd(reservationReader.getReservFinishDate());

				reservations.add(reservation);
			}
		}

		reservationDAO.insertBatch(reservations, false);

		log.debug("Processing reservations finished");
	}

	/**
	 * Process queues. It reads queues data from the stream and adds them to the
	 * database.
	 * 
	 * @param is
	 *            input stream with queues data
	 * @param skipHeaders
	 *            determines whether first line of input stream should be
	 *            skipped
	 * @throws SQLException
	 *             the SQL exception
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	private void processQueues(InputStream is, boolean skipHeaders)
			throws SQLException, IOException {

		log.debug("Processing queues");

		QueuesDAO queuesDAO = factory.createQueues();

		Queues queues = null;

		List<Queues> queuesList = new ArrayList<Queues>();

		RawResourcesReader queuesReader = new RawResourcesReader(
				getStatsParser(is, skipHeaders));

		while (queuesReader.next()) {

			queues = new Queues();

			queues.setName("default");
			queues.setFkResourceId(resourceID.get(queuesReader
					.getResourceName()));
			queues.setAvgLength(queuesReader.getQueueLength());

			queuesList.add(queues);
		}

		queuesDAO.insertBatch(queuesList, false);

		log.debug("Processing queues finished");
	}

	/**
	 * Process resource all timeline. It reads allocation timeline data from the
	 * stream and adds them to the database.
	 * 
	 * @param is
	 *            input stream with allocation timeline data
	 * @param skipHeaders
	 *            determines whether first line of input stream should be
	 *            skipped
	 * @throws SQLException
	 *             the SQL exception
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	private void processResourceAllTimeline(InputStream is, boolean skipHeaders)
			throws SQLException, IOException {

		log.debug("Processing resource allocation timelines");

		ResourceAllTimelineDAO resourceAllTimelineDAO = factory
				.createResourceAllTimeline();

		ResourceAllTimeline resourceAllTimeline = null;

		List<ResourceAllTimeline> resourceAllTimelines = new ArrayList<ResourceAllTimeline>();

		UsageReader resourceAllTimelineReader = new UsageReader(getStatsParser(
				is, skipHeaders));

		while (resourceAllTimelineReader.next()) {

			resourceAllTimeline = new ResourceAllTimeline();

			resourceAllTimeline.setFkResourceId(resourceID
					.get(resourceAllTimelineReader.getName()));
			resourceAllTimeline.setTimestamp(resourceAllTimelineReader
					.getTimeStamp());
			resourceAllTimeline.setUsage(resourceAllTimelineReader.getUsage());

			resourceAllTimelines.add(resourceAllTimeline);
		}

		resourceAllTimelineDAO.insertBatch(resourceAllTimelines, false);

		log.debug("Processing resource allocation timelines finished");
	}

	/**
	 * Process resource en timeline. It reads energy timeline data from the
	 * stream and adds them to the database.
	 * 
	 * @param is
	 *            input stream with energy timeline data
	 * @param skipHeaders
	 *            determines whether first line of input stream should be
	 *            skipped
	 * @throws SQLException
	 *             the SQL exception
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	private void processResourceEnTimeline(InputStream is, boolean skipHeaders)
			throws SQLException, IOException {

		log.debug("Processing resource energy timelines");

		ResourceEnTimelineDAO resourceEnTimelineDAO = factory
				.createResourceEnTimeline();

		ResourceEnTimeline resourceEnTimeline = null;

		List<ResourceEnTimeline> resourceEnTimelines = new ArrayList<ResourceEnTimeline>();

		UsageReader resourceEnTimelineReader = new UsageReader(getStatsParser(
				is, skipHeaders));

		while (resourceEnTimelineReader.next()) {

			resourceEnTimeline = new ResourceEnTimeline();

			resourceEnTimeline.setFkResourceId(resourceID
					.get(resourceEnTimelineReader.getName()));
			resourceEnTimeline.setTimestamp(resourceEnTimelineReader
					.getTimeStamp());
			resourceEnTimeline.setUsage(resourceEnTimelineReader.getUsage());

			resourceEnTimelines.add(resourceEnTimeline);
		}

		resourceEnTimelineDAO.insertBatch(resourceEnTimelines, false);

		log.debug("Processing resource energy finished");
	}

	/**
	 * Process network. It reads network data from the stream and adds them to
	 * the database.
	 * 
	 * @param is
	 *            input stream with network data
	 * @param skipHeaders
	 *            determines whether first line of input stream should be
	 *            skipped
	 * @throws SQLException
	 *             the SQL exception
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	private void processNetwork(InputStream is, boolean skipHeaders)
			throws SQLException, IOException {

		log.debug("Processing network");

		NetworkReader networkReader = new NetworkReader(getStatsParser(is,
				skipHeaders));

		processNetworkLinks(networkReader);
		processNetworkReservations(networkReader);

		log.debug("Processing network finished");
	}

	/**
	 * Processes network links. Called from processNetwork (only).
	 * 
	 * @param networkReader
	 *            the network reader
	 * @throws SQLException
	 *             the SQL exception
	 */
	private void processNetworkLinks(NetworkReader networkReader)
			throws SQLException {

		log.debug("Processing network links");

		ResourceDAO resourceDAO = factory.createResource();
		UnitDAO unitDAO = factory.createUnit();
		UnitDetailsDAO unitDetailsDAO = factory.createUnitDetails();

		Resource network = new Resource();

		network.setExpId(experiment.getPK());
		network.setName("network");
		network.setType("network");

		resourceDAO.insert(network, false);

		List<Unit> units = new ArrayList<Unit>();

		while (networkReader.next()) {

			if ("reservation ID".equals(networkReader.getFirstColumnString()))
				break;

			Unit unit = new Unit();
			UnitDetails baudRate = new UnitDetails();
			UnitDetails delay = new UnitDetails();
			UnitDetails sizeOfTransmittedData = new UnitDetails();

			unit.setName(networkReader.getNetworkLinks().getLinkName());
			unit.setAmount(1);
			unit.setResourceId(network.getPK());
			unit.setType("network");
			unit.setAllocationLoad(networkReader.getNetworkLinks().getLoad());
			unit.setReservationLoad(networkReader.getNetworkLinks()
					.getResevationLoad());

			baudRate.setName("baud rate");
			baudRate.setValue(networkReader.getNetworkLinks().getBaudRate());
			unit.addUnitDetails(baudRate);

			delay.setName("delay");
			delay.setValue(networkReader.getNetworkLinks().getDelay());
			unit.addUnitDetails(delay);

			sizeOfTransmittedData.setName("size of transmitted data");
			sizeOfTransmittedData.setValue(networkReader.getNetworkLinks()
					.getSizeOfTransmittedData());
			unit.addUnitDetails(sizeOfTransmittedData);

			units.add(unit);

		}

		unitDAO.insertBatch(units, false);

		List<UnitDetails> accUnitDetails = new ArrayList<UnitDetails>();

		for (Unit u : units) {
			unitID.put(u.getName(), u.getPK());
			u.getUnitDetailsOnList(accUnitDetails);
		}

		unitDetailsDAO.insertBatch(accUnitDetails, false);

	}

	/**
	 * Processes network reservations. Called from processNetwork (only).
	 * 
	 * @param networkReader
	 *            the network reader
	 * @throws SQLException
	 *             the SQL exception
	 */
	private void processNetworkReservations(NetworkReader networkReader)
			throws SQLException {

		log.debug("Processing network reservations");

		ReservationDAO reservationDAO = factory.createReservation();
		ReservationDetailsDAO reservationDetailsDAO = factory
				.createReservationDetails();

		Reservation reservation = null;
		List<Reservation> reservations = new ArrayList<Reservation>();

		while (networkReader.next()) {

			if ("JobID".equals(networkReader.getFirstColumnString()))
				break;

			String unitNamesString = networkReader.getNetworkReservations()
					.getLinks();
			int endIndex = Math.max(1, unitNamesString.length() - 1);
			unitNamesString = unitNamesString.substring(1, endIndex);
			String[] unitNames = unitNamesString.split(", ");

			for (String unitName : unitNames) {

				reservation = new Reservation();
				reservation.setPK(networkReader.getNetworkReservations()
						.getReservaionId());
				reservation.setUnitId(unitID.get(unitName));
				reservation.setBegin(networkReader.getNetworkReservations()
						.getReservStartDate());
				reservation.setEnd(networkReader.getNetworkReservations()
						.getReservFinishDate());

				ReservationDetails bandwidth = new ReservationDetails();
				bandwidth.setName("bandwidth");
				bandwidth.setValue(networkReader.getNetworkReservations()
						.getBandwidth());
				reservation.addReservationDetails(bandwidth);

				ReservationDetails load = new ReservationDetails();
				load.setName("load");
				load.setValue(networkReader.getNetworkReservations().getLoad());
				reservation.addReservationDetails(load);

				ReservationDetails sizeOfTransmittedData = new ReservationDetails();
				sizeOfTransmittedData.setName("size of transmitted data");
				sizeOfTransmittedData.setValue(networkReader
						.getNetworkReservations().getSizeOfTransmittedData());
				reservation.addReservationDetails(sizeOfTransmittedData);

				ReservationDetails reservationStatus = new ReservationDetails();
				reservationStatus.setName("status");
				reservationStatus.setValue(networkReader
						.getNetworkReservations().getReservationStatus());
				reservation.addReservationDetails(reservationStatus);

				reservations.add(reservation);
			}
		}

		while (networkReader.next()) {
			Integer jobId = networkReader.getNetworkReservationsTasks()
					.getJobID();
			Integer taskId = networkReader.getNetworkReservationsTasks()
					.getTaskID();
			Integer reservationId = networkReader.getNetworkReservationsTasks()
					.getReservationID();

			Pair<Integer, Integer> jobTaskID = new Pair<Integer, Integer>(
					jobId, taskId);

			networkTaskID.put(reservationId, taskID.get(jobTaskID));
		}

		for (Reservation r : reservations) {
			Integer taskId = networkTaskID.get(r.getPK());
			r.setPK(null);
			r.setTaskId(taskId);
		}

		reservationDAO.insertBatch(reservations, false);

		List<ReservationDetails> accReservationDetails = new ArrayList<ReservationDetails>();

		for (Reservation r : reservations) {
			r.getReservationDetailsOnList(accReservationDetails);
		}

		reservationDetailsDAO.insertBatch(accReservationDetails, false);

		log.debug("Processing network reservations finished");
	}

	/**
	 * Processes network topology. It reads network topology from the stream and
	 * adds them to the database.
	 * 
	 * @param is
	 *            input stream with network topology data
	 * @throws SQLException
	 *             the SQL exception
	 */
	private void processNetworkTopology(InputStream is) throws SQLException {

		log.debug("Processing network topology");

		BlobPropertyDAO blobPropertyDAO = factory.createBlobProperty();
		BlobProperty blobProperty = new BlobProperty();
		blobProperty.setConfigId(configuration.getId());
		blobProperty.setName("Network topology");
		blobProperty.setValue(IOUtils.toByteArray(inputStreams
				.get("networkTopologyInputStream")));
		blobPropertyDAO.insert(blobProperty, false);

		log.debug("Processing network topology finished");
	}
}